﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Linq;
using System.ServiceModel;
using System.Threading;
using VVIInvestment.RusGuard.DAL.Entities.Entity;
using VVIInvestment.RusGuard.DAL.Entities.Entity.ACS.AccessLevels;
using VVIInvestment.RusGuard.DAL.Entities.Entity.ACS.AccessLevels.AcsAccesPoints;
using VVIInvestment.RusGuard.DAL.Entities.Entity.ACS.AcsKeys;
using VVIInvestment.RusGuard.DAL.Entities.Entity.ACS.Employees;
using VVIInvestment.RusGuard.DAL.Entities.Entity.Log;
using VVIInvestment.RusGuard.DAL.Entities.Exceptions;
using VVIInvestment.RusGuard.Net.Services;
using VVIInvestment.RusGuard.Net.Services.Entities;
using VVIInvestment.RusGuard.Net.Services.Wcf;

/*
 * Данный класс содержит примеры для интеграции со СКУД системы RusGuard
 * 
 * Для взаимодействия с внешним миром система выставляет WCF сервис. У сервиса всегда включен режим безопасности TransportWithMessageCredential,
 * соответственно всегда используется HTTPS, а также аутентификация и авторизация пользователя на основе его логина и пароля. Логин и пароль по умолчанию
 * "Admin" и "" соответственно, но, естественно, может меняться при настройке системы + пользователю могут задаваться различные права, которые 
 * проверяются при его авторизации.
 * 
 * Поэтому:
 * a) всегда надо быть готовым к тому, что вызов любого метода сервиса выкинет исключение MessageSecurityException в случае, 
 *    если пользователь не прошел аутентификацию, а также SecurityAccessDeniedException если пользователь не прошел авторизацию для 
 *    выполнения того или иного действия.
 * б) для взаимодействия по HTTPS чаще всего на сервере генерируется самозаверенный сертификат, отсутствующий в списке доверенных сертификатов клиента,
 *    что приводит к тому, что клиенты не могут присоединиться к серверу, т.к. сертификат на клиенте не проходит валидацию. Это поведение можно переопределить.
 *    Для .Net это делается с помощью ServicePointManager.ServerCertificateValidationCallback
 * 
 * WCF сервис выставляет наружу две конечные точки:
 * 1. Точка с контрактом ILNetworkService предназначена для извлечения из системы различных данных
 * 2. Точка с контрактом ILNetworkConfigurationService предназначена для конфигурирования системы
 * 
 * Все методы данного класса носят сугубо демонстрационный характер. Никакой обработки серверных исключений нет.
 * Серверные исключения перехватываются с одной единственной целью - продемонстрировать, какие они вообще есть с комментариями о причинах возникновения.
 * 
 * Методы сгруппированы в регионы, каждый из которых содержит набор методов и комментариев, относящихся к реализации одной из задач интеграции.
 * 
 * Для вызовов к удаленному сервису используются наши обертки вокруг WCF прокси.
 * 
 * На жестком диске В папке Libs проекта находятся *.dll, на который ссылается данный проект и которые содержат контракты WCF сервисов,
 * данные, которыми обмениваются клиент и сервис и специфичные исключения, которые сервис может выбрасывать в ответ на запросы клиентов.
 * 
 * Ко всем *.dll приложены *.xml файлы с комментариями, соответственно для любого метода/класса/свойства доступны их контракт с комментариями, 
 * отображаемые IntelliSense.По это причине контракты данных, учавствующих во взаимодействии клиента и сервера, отдельно не описываются. 
 * Однако, если объект содержит данные других объектов системы СКУД, то об этом упоминается в комментарии. 
 * Например, "Данные каждого сотрудника содержат назначенные ему уровни доступа, ключи, табельный номер"
 * 
 */

namespace RusGuardAcsIntegrationSample
{
    internal class Program
    {
        private const string ServiceUrl = "https://10.0.0.25/LNetworkServer/LNetworkService.svc";

        private static readonly ChannelContainer<ILNetworkService> NetworkService = new ChannelContainer<ILNetworkService>(
            "Admin",
            "",
            WCFBindingFactory.CreateUnlimitedBasicHttpBinding(BasicHttpSecurityMode.TransportWithMessageCredential));

        private static readonly ChannelContainer<ILNetworkConfigurationService> NetworkCnfgService = new ChannelContainer<ILNetworkConfigurationService>(
            "Admin",
            "",
            WCFBindingFactory.CreateUnlimitedBasicHttpBinding(BasicHttpSecurityMode.TransportWithMessageCredential));

        private static void Main(string[] args)
        {
            ServicePointManager.ServerCertificateValidationCallback = (a, b, c, d) => true;
           
            var t = GetAcsEmployeeGroups();

            Console.WriteLine(t.Length);

            Console.ReadKey();
        }

        #region Получить уровни доступа

        private static AcsAccessLevelSlimInfo[] GetAcsAccessLevels()
        {
            //Следует учитывать, что у уровней доступа есть свойство IsRemoved которое означает, что хотя он и существует в БД, 
            //но помечен как удаленный и все операции с ним будут приводить к выбрасыванию исключения. 
            //В нашем случае такие уровни доступа просто не обрабатываем.
            return NetworkService.InvokeFunc(_ => _.GetAcsAccessLevelsSlimInfo(), ServiceUrl).Where(_ => !_.IsRemoved).ToArray();
        }

        #endregion

        #region Работа с общими данными сотрудников/посетителей, назначение уровней доступа

        /*
         * ВАЖНАЯ информация по работе с сотрудниками
         * 
         * 1. Сотрудники не могут существовать сами по себе, они обязательно должны 
         *    входить в какую-либо группу сотрудников. Группы сотрудников имеют древовидную структуру, дочерние группы содержатся в свойстве EmployeeGroups группы.
         *    Может быть несколько корневых групп. У каждой группы есть имя и содержащиеся в ней сотрудники. Сотрудник в каждый момент времени может находиться 
         *    только в одной группе.
         *    
         * 2. Если требуется заводить в систему гостей, то следует учитывать, что специализированной группы именно для посетителей в системе нет, 
         *    поэтому следует использовать следующий алгоритм. 
         *    Загружаем список групп сотрудников и ищем корневую с именем "Посетители", если не нашли, то создаем и далее добавляем в нее сотрудников.
         *    При этом надо учитывать, что оператор через АРМ системы RusGuard эту группу сотрудников может в любой момент удалить. В этом случае при попытке 
         *    добавить в нее сотрудников будет выброшено исключение, обработкой которого должно являться повторное создание группы сотрудников для посетителей.
         *    
         * 3. При назначении сотрудникам уровней доступа необходимо понимать назначение флага IsAccessLevelsInherited, который указывает, действуют ли для сотрудника
         *    уровни доступа, назначенные группе сотрудников, в которой он состоит, или же для сотрудника действуют уровни доступа, назначенные ему непосредственно.
         *    В связи с вышеописанным, существует два способа назначения уровней доступа сотруднику. Первый способ заключается в установке флага IsAccessLevelsInherited в true для сотрудника.
         *    Второй способ заключается в непосредственном назначении уровней доступа сотруднику, при этом флаг IsAccessLevelsInherited  должен быть снят
         *    
         */



        /// <summary>
        /// Получить все группы сотрудников.
        /// </summary>
        /// <returns></returns>
        private static AcsEmployeeGroup[] GetAcsEmployeeGroups()
        {
            //Следует учитывать, что у групп сотрудников есть свойство IsRemoved которое означает, что хотя она и существует в БД, 
            //но помечена как удаленная и все операции с ней будут приводить к выбрасыванию исключения. 
            //В нашем случае такие группы просто не обрабатываем.
            return NetworkService.InvokeFunc(_ => _.GetAcsEmployeeGroups(), ServiceUrl).Where(_ => !_.IsRemoved).ToArray();
        }

        /// <summary>
        /// Пример получения группы сотрудников для посетителей
        /// </summary>
        /// <returns></returns>
        public static AcsEmployeeGroup GetGuestEmployeeGroup()
        {
            AcsEmployeeGroup guestEmployeeGroup = NetworkService.InvokeFunc(_ => _.GetAcsEmployeeGroups(), ServiceUrl).FirstOrDefault(_ => !_.IsRemoved && _.Name == "Посетители");

            try
            {
                if (guestEmployeeGroup == null)
                {
                    guestEmployeeGroup = NetworkCnfgService.InvokeFunc(_ => _.AddAcsEmployeeGroup(null, "Посетители", ""), ServiceUrl);
                }
                return guestEmployeeGroup;

            }
            //Наименование создаваемой группы равно null или является пустой строкой
            catch (FaultException<ArgumentException>)
            {
                throw;
            }
            //Наименование создаваемой группы содержит более 255 символов
            catch (FaultException<ArgumentOutOfRangeException>)
            {
                throw;
            }
        }


        /// <summary>
        /// Получить краткую информацию о сотрудниках в группе. Данные каждого сотрудника содержат назначенные ему уровни доступа, ключи, табельный номер
        /// </summary>
        /// <param name="groupId"></param>
        private static AcsEmployeeSlim[] GetAcsEmployeesInGroup(Guid groupId)
        {
            try
            {
                // 1. Следует учитывать, что у сотрудников есть свойство IsRemoved которое означает, что хотя он и существует в БД, 
                //но помечен как удаленный и все операции с ним будут приводить к выбрасыванию исключения. 
                //В нашем случае такие группы просто не обрабатываем.
                // 2. Сведения о сотрудниках содержат только минимальное необходимые данный, для получения расширенной информации по сотруднику
                //необходимо воспользоваться методом GetAcsEmployee
                return  NetworkService.InvokeFunc(_ => _.GetAcsEmployeesByGroup(groupId), ServiceUrl).Where(_ => !_.IsRemoved).ToArray();
            }
            //Исключение возникает в случае, если группа сотрудников отсутствует в БД или ее свойство IsRemoved установлено в true
            catch (FaultException<DataNotFoundException>)
            {
                throw;
            }
        }

        /// <summary>
        /// Получить полные данные о сотрудниках по табельным номерам
        /// </summary>
        /// <returns></returns>
        public static AcsEmployeeFull[] GetAcsEmployeesByTableNumbers(int[] tableNumbers)
        {
            //полная информация содержит паспортные данные, состояние сотрудника(удален), сведения о блокировании - разблокировании сотрудника и т.д
            return NetworkService.InvokeFunc(_ => _.GetAcsEmployeesByTableNumbers(tableNumbers), ServiceUrl);
        }

        /// <summary>
        /// Получить полные данные о сотруднике
        /// </summary>
        /// <returns></returns>
        public static AcsEmployeeFull GetAcsEmployee(Guid id)
        {
            //полная информация содержит паспортные данные, состояние сотрудника(удален), сведения о блокировании - разблокировании сотрудника и т.д
            return NetworkService.InvokeFunc(_ => _.GetAcsEmployee(id), ServiceUrl);
        }
        /// <summary>
        /// Получение всех уровней доступа 
        /// </summary>
        /// <returns></returns>
        public static AcsAccessLevelSlimInfo[] GetAcsAccessLevelsSlimInfo()
        {

            return NetworkService.InvokeFunc(_ => _.GetAcsAccessLevelsSlimInfo(), ServiceUrl).ToArray();
        }

        /// <summary>
        /// Добавление сотрудников в группу
        /// </summary>
        /// <param name="employeeGroupID"></param>
        /// <param name="data"></param>
        /// <returns></returns>
        public static AcsEmployeeSlim AddAcsEmployee(Guid employeeGroupID, AcsEmployeeSaveData data)
        {

            try
            {
                return NetworkCnfgService.InvokeFunc(_ => _.AddAcsEmployee(employeeGroupID, data), ServiceUrl);
            }
            //Параметр data равено null
            catch (FaultException<ArgumentNullException>)
            {
                throw;
            }
            // 1. Поля FirstName, LastName, SecondName параметра data: имеют длину более 255 символов.
            // 2. Поле PassportIssue имеет длину более 50 символов.
            // 3. Поле PassportNumber имеет длину более 100 символов.
            catch (FaultException<ArgumentOutOfRangeException>)
            {
                throw;
            }
            //Группа сотрудников, переданная в параметре employeeGroupID отсутствует в БД или помечена как удаленная.
            //Должность, переданная в поле EmployeePositionID параметра data отсутствует в БД или помечена как удаленная.
            catch (FaultException<DataNotFoundException>)
            {
                throw;
            }
            //Табельный номер, переданный в поле Number , уже назначен другому сотруднику
            catch (FaultException<DataAlreadyExistsException>)
            {
                throw;
            }
            //должно быть задано как минимум одно из следующих полей в параметре data: FirstName, LastName, SecondName
            catch (FaultException<InvalidOperationException>)
            {
                throw;
            }
        }     


        /// <summary>
        /// Сохранить данные сотрудника.
        /// </summary>
        /// <returns></returns>
        public static void SaveAcsEmployee(Guid id, AcsEmployeeSaveData data)
        {

            try
            {
                NetworkCnfgService.InvokeAction(_ => _.SaveAcsEmployee(id, data), ServiceUrl);
            }
            //Параметр data равено null
            catch (FaultException<ArgumentNullException>)
            {
                throw;
            }
            // 1. Поля FirstName, LastName, SecondName параметра data: имеют длину более 255 символов.
            // 2. Поле PassportIssue имеет длину более 50 символов.
            // 3. Поле PassportNumber имеет длину более 100 символов.
            catch (FaultException<ArgumentOutOfRangeException>)
            {
                throw;
            }
            //Группа сотрудников, переданная в параметре employeeGroupID отсутствует в БД или помечена как удаленная(IsRemoved = true).
            //Должность, переданная в поле EmployeePositionID параметра data отсутствует в БД или помечена как удаленная(IsRemoved = true).
            catch (FaultException<DataNotFoundException>)
            {
                throw;
            }
            //Табельный номер, переданный в поле Number , уже назначен другому сотруднику
            catch (FaultException<DataAlreadyExistsException>)
            {
                throw;
            }
            //должно быть задано как минимум одно из следующих полей в параметре data: FirstName, LastName, SecondName
            catch (FaultException<InvalidOperationException>)
            {
                throw;
            }
        }

        /// <summary>
        /// Удаление сотрудника
        /// </summary>
        /// <param name="id"></param>
        public static void RemoveAcsEmployee(Guid id)
        {
            NetworkCnfgService.InvokeAction(_ => _.RemoveAcsEmployee(id), ServiceUrl);
        }

        /// <summary>
        /// Добавить фотографию сотруднику, удалить фотографию у сотрудника
        /// </summary>
        /// <param name="employeeId"></param>
        /// <param name="photoNumber"></param>
        /// <param name="data"></param>
        public static void SetAcsEmployeePhoto(Guid employeeId, int photoNumber, byte[] data)
        {

            try
            {
                //photoNumber - параметр, указывающий номер(поизицию фотографии), нумерация начинается с еденицы.
                //В настройках сервисов есть ограничение на объем передаваемых данных в параметре data, 
                //data не должно превышать 4 Mb
                //Для удаления фотографии с позицией, указанной в photoNumber, в параметре data следует указать null
                NetworkCnfgService.InvokeAction(_ => _.SetAcsEmployeePhoto(employeeId, photoNumber, data), ServiceUrl);
            }
            //сотрудник, переданная в параметре employeeId отсутствует в БД или помечен как удаленный(IsRemoved = true).
            catch (FaultException<DataNotFoundException>)
            {
                throw;
            }

        }

        /// <summary>
        /// Назначить уровни доступа оструднику
        /// </summary>
        /// <param name="employeeID"></param>
        /// <param name="accessLevelIDs"></param>
        public static void AddAccessLevelsToEmployee(Guid employeeID, Guid[] accessLevelIDs)
        {
            try
            {
                NetworkCnfgService.InvokeAction(_ => _.AddAccessLevelsToEmployee(employeeID, accessLevelIDs), ServiceUrl);
            }
            //Исключение возникает в случае, если отрудник отсутствует в БД или его свойство IsRemoved установлено в true
            //и/или какой-либо уровень доступа отсутствуют в БД или его свойство IsRemoved установлено в true
            catch (FaultException<DataNotFoundException>)
            {
                throw;
            }
            //Исключение возникает в случае, если в списке идентификаторов уровней доступа существует хотя бы один уровень доступа
            //уже назначенный сотруднику
            catch (FaultException<DataAlreadyExistsException>)
            {
                throw;
            }
            //Исключение возникает в случае, если происходит попытка добавить уровни доступа сотруднику, у которого флаг IsAccessLevelsInherited
            //выставлен в true
            catch (FaultException<InvalidOperationException>)
            {
                throw;
            }
        }

        /// <summary>
        /// Удалить уровни доступа у сотрудника
        /// </summary>
        /// <param name="employeeID"></param>
        /// <param name="accessLevelIDs"></param>
        public static void RemoveAccessLevelFromEmployee(Guid employeeID, Guid[] accessLevelIDs)
        {
            try
            {

                NetworkCnfgService.InvokeAction(_ => _.RemoveAccessLevelFromEmployee(employeeID, accessLevelIDs), ServiceUrl);
            }
            //Уровень досутпа отсутствует в БД или помечен как удаленный
            catch (FaultException<DataNotFoundException>)
            {
                throw;
            }
        }

        /// <summary>
        /// Установка-снятие флага IsAccessLevelsInherited
        /// </summary>
        /// <param name="employeeID"></param>
        /// <param name="isUseParentAccessLevel"></param>
        public static void SetUseEmployeeParentAccessLevel(Guid employeeID, bool isUseParentAccessLevel)
        {
            try
            {
                NetworkCnfgService.InvokeAction(_ => _.SetUseEmployeeParentAccessLevel(employeeID, isUseParentAccessLevel), ServiceUrl);
            }
            //Исключение возникает в случае, если отрудник отсутствует в БД или ее свойство IsRemoved установлено в true
            catch (FaultException<DataNotFoundException>)
            {
                throw;
            }
        }

        #endregion

        #region Блокировка - разблокировка сотрудников

        /*
         * 1. Сотрудника можно заблокировать/разблокировать (т.е. запретить/разрешить ему перемещаться по объекту). При этом все ключи, которые ему были назначены
         * ранее, так и остаются за ним закреплены и попытка задать эти ключи другому сотруднику вызовет исключение (см. ниже пример назначения ключа сотруднику)
         * 2. Узнать состояние блокировки сотрудника можно, воспользовавшись методом GetAcsEmployee
         */

        /// <summary>
        /// Заблокировать или разблокировать сотрудников.
        /// </summary>
        /// <param name="ids"></param>
        /// <param name="isLocked"></param>
        public static void LockAcsEmployee(Guid[] ids, bool isLocked)
        {
            try
            {
                NetworkCnfgService.InvokeAction(_ => _.LockAcsEmployee(ids, isLocked), ServiceUrl);
                
            }
            //параметр ids пустой или null
            catch (FaultException<ArgumentException>)
            {
                throw;
            }
            //По крайне мере один из сотрудников, чей идентификатор передан в ids отсутствует в базе
            //или IsRemoved = true
            catch (FaultException<DataNotFoundException>)
            {
                throw;
            }
        }

        #endregion

        #region Работа с ключами

        /* ВАЖНАЯ информация по работе с ключами
         * 1. У сотрудника может быть назначено только два ключа с порядковыми номерами 1 и 2
         * 
         * 2. Если дата начала и окончания действия ключа не указана(StartDate и EndDate соответственно) то считается, что ключ выдается
         * на все время(действует бесконечно). Если указана только дата начала, то ключ действует с этой даты и до бесконечности.
         * Если указана только дата окончания действия, то ключ действует с текущего момента до момента наступления даты окончания действия ключа.
         * 
         * 3. При попытке назначить ключ сотруднику может возникнуть ситтуация, когда ключ уже назначен другому сотруднику.
         * В этом случае допустимо воспользоваться методом ForceAssignAcsKeyForEmployee и переназначить ключ. Сценарий назначения ключа сотруднику в общем случает такой:
         * Попытка назначить ключ посредством метода AssignAcsKeyForEmployee, если возникает исключение AssignmentAcsKeyException с типом ошибки AssignmentAcsKeyErrorType.AcsKeyAlreadyAssignedToAnotherEmployee,
         * то принимается решение об продолжении операции назначения ключа посредством метода ForceAssignAcsKeyForEmployee, или отказе от операции
         */

                
        
        /// <summary>
        /// Забрать ключ у сотрудника
        /// </summary>
        /// <param name="employeeId"></param>
        /// <param name="indexNumber"></param>
        /// <returns></returns>
        public static void RemoveKeyFromEmployee(Guid employeeId, int indexNumber)
        {

            NetworkCnfgService.InvokeFunc(_ => _.AssignAcsKeyForEmployee(employeeId, indexNumber, null), ServiceUrl);

        }

        /// <summary>
        /// Привязать ключ к сотруднику с трактовкой операции как недопустимой, если ключ уже присвоен другому сотруднику.
        /// </summary>
        /// <param name="employeeId"></param>
        /// <param name="indexNumber"></param>
        /// <param name="keyData"></param>
        /// <returns></returns>
        public static AcsKeyInfo AssignAcsKeyForEmployee(Guid employeeId, int indexNumber, AcsKeySaveData keyData)
        {
            try
            {
                //если в параметре keyData указать null, то ключ у сотрудника будет отнят
                return NetworkCnfgService.InvokeFunc(_ => _.AssignAcsKeyForEmployee(employeeId, indexNumber, keyData), ServiceUrl);
            }
            //Сотрудник с идентификатором, переданным в employeeId отсутствует в БД или помечен как удаленный(IsRemoved = true).
            catch (FaultException<DataNotFoundException>)
            {
                throw;
            }
            //В данных ключа, переданных в keyData дата начала и дата окончания действия ключа (StartDate и EndDate соответственно) заданы, но дата начала больше даты окончания.
            catch (FaultException<InvalidOperationException>)
            {
                throw;
            }
            // 1. Ключ уже назначен другому сотруднику( ex.Detail.ErrorType == AssignmentAcsKeyErrorType.AcsKeyAlreadyAssignedToAnotherEmployee)
            // 2. Ключ уже назначен данному сотруднику под другим порядковым номером(ex.Detail.ErrorType == AssignmentAcsKeyErrorType.AcsKeyAlreadyAssignedToCurrentEmployee).
            catch (FaultException<AssignmentAcsKeyException> ex)
            {   
                throw;
            }
        }

        /// <summary>
        /// Привязать ключ к сотруднику. Если ключ уже присвоен другому сотруднику, то он у него будет сброшен.
        /// </summary>
        /// <param name="employeeId">Идентификатор сотрудника.</param>
        /// <param name="indexNumber">Номер ключа в списке ключей сотрудника.</param>
        /// <param name="keyData">Данные ключа. null, если ключ следует сбросить</param>
        public static AcsKeyInfo ForceAssignAcsKeyForEmployee(Guid employeeId, int indexNumber, AcsKeySaveData keyData)
        {
            try
            {
                return NetworkCnfgService.InvokeFunc(_ => _.ForceAssignAcsKeyForEmployee(employeeId, indexNumber, keyData), ServiceUrl);
            }
            //Сотрудник с идентификатором, переданным в employeeId отсутствует в БД или помечен как удаленный(IsRemoved = true).
            catch (FaultException<DataNotFoundException>)
            {
                throw;
            }
            //В данных ключа, переданных в keyData дата начала и дата окончания действия ключа (StartDate и EndDate соответственно) заданы, но дата начала больше даты окончания.
            catch (FaultException<InvalidOperationException>)
            {
                throw;
            }
            // Ключ уже назначен данному сотруднику под другим порядковым номером.
            catch (FaultException<AssignmentAcsKeyException>)
            {
                throw;
            }
        }


        #endregion

        #region Работа с событиями

        /*  ВАЖНАЯ информация по работе с событиями
         *  1. Для постоянного мониторинга событий, происходящих на сервере, фильтр по дате/времени использовать нельзя (время сервера и клиента могут различаться)
         *  Каждое событие имеет номер, являющийся идентификатор записи(строки) в БД.
         *  Идентификаторы инкриментируются, поэтому у позднего события идентификатор всегда больше, чем у раннего. В связи с чем сценарий мониторинга событий
         *  сводится к цикличному вызову метода GetLastEvent, чтобы получить идентификатор последнего события, удовлетворяющего фильтру, 
         *  принятие его в качестве максимального значения идентификатора и затем цикличный вызов GetEvents, 
         *  поиск нового максимального идентификатора записи среди списка полученных событий, чтобы на последующей итерации вызвать метод, 
         *  указав в качестве параметра "с какой записи следует вычитка события" это идентификатор.
         *  Следует устанавливать задержку между итерациями опроса. Величина задержки должна быть адекватна задаче, решаемой с помощью мониторинга событий.
         *  Безостановочно вызывать серверный метод не стоит.
         *  
         *  2. Если же требуется просто снять данные за определенный период времени, то следует вызвать GetEvents, указав в параметрах нужные даты.
         *  
         *  3. ВАЖНО. Очень нежелательно (особенно это касается постоянного мониторинга) получать события, используя в качестве одного из фильтров идентификаторы точек доступа, 
         *  т.к. на реальном объекте их может быть не одна тысяча, что существенно замедляет выборку. Следует получить список точек доступа, после чего 
         *  сопоставлять идентификаторы устройств в событиях с идентификаторами полученных точек доступа. Следует учитывать, что список точек доступа хоть и достаточно редко 
         *  (часто на этапе пуско-наладки), но меняется, т.к. точки прохода периодически монтируются/демонтируются.
         */

        /// <summary>
        /// Пример постоянного постоянного мониторинга событий для всех или определенных точек доступа по фильтру событий входа и выхода
        /// </summary>
        /// <param name="accessPointsIds">null, если события необходимо получать для всех точек доступа</param>
        public static void TrackEvents(Guid[] accessPointsIds)
        {
            LogMsgSubType[] filter = new []
                                         {
                                             //вход
                                             LogMsgSubType.AccessPointEntryByKey,
                                             LogMsgSubType.AccessPointFirstPersonEntry,
                                             LogMsgSubType.AccessPointSecondPersonEntry,
                                             //выход
                                             LogMsgSubType.AccessPointExitByKey,
                                             LogMsgSubType.AccessPointExitByCardReceiver,
                                             LogMsgSubType.AccessPointFirstPersonExit,
                                             LogMsgSubType.AccessPointFirstPersonExitByCardReceiver,
                                             LogMsgSubType.AccessPointSecondPersonExit,
                                             LogMsgSubType.AccessPointSecondPersonExitByCardReceiver
                                         };

            //Metоды возвращают данные о событиях в типе LogData. Список событий непосредственно представлен в массиве  Messages
            //Каждый элемент этого массива представляет событие, где LogMessageSubType - расшифровывает что произошло в точке доступа, идентификатор
            //которой указан в поле DriverID а ее наименование в DriverName
            long counter = -1;

            //Считываем последнее событие, удовлетворяющее фильтру, принимаем его идентификатор за максимальный.
            while (counter == -1)
            {
                var logData = NetworkService.InvokeFunc(_ => _.GetLastEvent(null, filter, accessPointsIds, LogSubjectType.None), ServiceUrl);

                if (logData != null && logData.Count > 0)
                {
                    counter = logData.Messages.Max(x => x.Id);
                }
            }

            //Вычитываем события от максимального идентификатора, обновляя его значение по списку полученных событий
            while (true)
            {
			    //для всех устройств
                var events = NetworkService.InvokeFunc(_ => _.GetEvents(counter, DateTime.MinValue, DateTime.MaxValue, null, filter, null, LogSubjectType.None, 0, 1000,
                                                                        LogMessageSortedColumn.DateTime, SortOrder.Ascending), ServiceUrl);
				//если необходим фильтр по устройствам, то раскомментировать нижеследующее
				//var events = NetworkService.InvokeFunc(_ => _.GetEventsByDeviceIDs(counter, DateTime.MinValue, DateTime.MaxValue, null, filter, accessPointsIds, null, LogSubjectType.None, 0, 1000,
                //                                                        LogMessageSortedColumn.DateTime, SortOrder.Ascending), ServiceUrl);

                if (events != null && events.Count > 0)
                {
                    var messagesEvents = events.Messages;
                    counter = events.Messages.Max(x => x.Id);
                    
                    //какие-то действия с событиями
                }
                Thread.Sleep(TimeSpan.FromSeconds(2));
            }
        }

        /// <summary>
        /// Пример получения событий для всех или определенных точек доступа по фильтру событий входа и выхода за январь
        /// </summary>
        /// <param name="accessPointsIds">null, если события необходимо получать для всех точек доступа</param>
        public static LogData GetEvents(Guid[] accessPointsIds)
        {
            LogMsgSubType[] filter = new[]
                                         {
                                             //вход
                                             LogMsgSubType.AccessPointEntryByKey,
                                             LogMsgSubType.AccessPointFirstPersonEntry,
                                             LogMsgSubType.AccessPointSecondPersonEntry,
                                             //выход
                                             LogMsgSubType.AccessPointExitByKey,
                                             LogMsgSubType.AccessPointExitByCardReceiver,
                                             LogMsgSubType.AccessPointFirstPersonExit,
                                             LogMsgSubType.AccessPointFirstPersonExitByCardReceiver,
                                             LogMsgSubType.AccessPointSecondPersonExit,
                                             LogMsgSubType.AccessPointSecondPersonExitByCardReceiver
                                         };

            var beginDate = new DateTime(2012, 1, 1);
            var endDate = new DateTime(2012, 1, 31, 23, 59, 59);
            var events = NetworkService.InvokeFunc(_ => _.GetEventsByDeviceIDs(0, beginDate, endDate, null, filter, accessPointsIds, null, LogSubjectType.None, 0, 1000,
                                                        LogMessageSortedColumn.DateTime, SortOrder.Ascending), ServiceUrl);
            return events;
        }

        private static AcsAccessPointDriverInfo[] GetAccessPoints()
        {
            return NetworkService.InvokeFunc(_ => _.GetAcsAccessPointDrivers(), ServiceUrl);
        }

        #endregion

        #region Получение драйверов устройств, посылка комманд

        /// <summary>
        /// Получить коллекцию всех драйверов устройств
        /// </summary>
        public static List<LDriverFullInfo> GetAllDrivers()
        {
            try
            {
                var resultDrivers = new List<LDriverFullInfo>();
                var networks = NetworkService.InvokeFunc(_ => _.GetAllNets(), ServiceUrl);
                for (int index = 0; index < networks.Length; index++)
                {
                    var lNetInfo = networks[index];
                    var servers = NetworkService.InvokeFunc(_ => _.GetNetServers(lNetInfo.Id), ServiceUrl);
                    for (int i = 0; i < servers.Length; i++)
                    {
                        var lServerInfo = servers[i];
                        var drivers = NetworkService.InvokeFunc(_ => _.GetServerDriversFullInfo(lServerInfo.Id), ServiceUrl);
                        resultDrivers.AddRange(drivers);
                    }
                }
                return resultDrivers;
            }
            // Идентификатор сервера устройств пустой
            catch (FaultException<ArgumentException> ex)
            {
                throw;
            }
        }

        /// <summary>
        /// Получение драйверов точек доступа, предварительно получив все дрйвера устройств. Лучше использовать GetAccessPointsDrivers2, т.к. надор типов точек доступа может расширяться.
        /// </summary>
        /// <returns></returns>
        public static IEnumerable<LDriverFullInfo>  GetAccessPointsDrivers1()
        {
             ///!!!!!!!!!!!!!Лучше использовать GetAccessPointsDrivers2, т.к. набор типов точек доступа может расширяться.!!!!!!!!!!!!!!!!!!!

            //список полных имен типов драйверов точек доступа
            string[] templatesAssemblyQualifiedName = new[]
                                                          {
                                                              "VVIInvestment.RusGuard.Driver.Acs.OneSidedDoorDriver, RusGuard.Driver.Acs, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null",
                                                              "VVIInvestment.RusGuard.Driver.Acs.TwoSidedDoorDriver, RusGuard.Driver.Acs, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null",
                                                              "VVIInvestment.RusGuard.Driver.Acs.TourniquetDriver, RusGuard.Driver.Acs, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null",
                                                              "VVIInvestment.RusGuard.Driver.Acs.GateDriver, RusGuard.Driver.Acs, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null"
                                                          };
            return GetAllDrivers().Where(x => templatesAssemblyQualifiedName.Any(name =>name.Equals(x.DriverType)));

        }


        /// <summary>
        /// Получение драйверов точек доступа, используя сервисный метод
        /// </summary>
        /// <returns></returns>
        public static AcsAccessPointDriverInfo[] GetAccessPointsDrivers2()
        {
            return NetworkService.InvokeFunc(_ => _.GetAcsAccessPointDrivers(), ServiceUrl);
        }


        /// <summary>
        /// Послать команду точке доступа
        /// </summary>
        /// <param name="connectionId"></param>
        /// <param name="driverId"></param>
        /// <param name="commandName"></param>
        public static void SendCommand(Guid connectionId, Guid driverId, string commandName)
        {

            //для точек доступа имена комманд варьируются

            //для односторонней и двусторонней двери допустимые commandName следующие:
            //Block(Блокировать ТД)
            //Unblock(Разблокировать ТД)
            //ArmGuard(Поставить датчики на охрану принудительно)
            //DisarmGuard(Снять датчики с охраны)
            //TurnOnSiren(Включить сирену)
            //TurnOffSiren(Выключить сирену)
            //LongOpen(Открыть надолго)
            //Open(Открыть)
            //Close(Закрыть)

            //для турникета допустимые commandName следующие:
            //Block(Блокировать ТД)
            //Unblock(Разблокировать ТД)
            //ArmGuard(Поставить датчики на охрану принудительно)
            //DisarmGuard(Снять датчики с охраны)
            //OpenEntry(Открыть на вход)
            //OpenExit(Открыть на выход)
            //Close(Закрыть)

            //для ворот допустимые commandName следующие:
            //Open(Открыть)
            //Close(Закрыть)
            //LongOpen(Открыть надолго)
            //Block(Блокировать ТД)
            //Unblock(Разблокировать ТД)
            //EmergencyOpen(Аварийное открытие)
            //ArmGuard(Поставить датчики на охрану принудительно)
            //DisarmGuard(Снять датчики с охраны)

            var operation = new DeviceCallMethodOperation(new Guid[0], null, driverId, commandName, new string[0]);

            NetworkService.InvokeAction(_ => _.Process(operation, connectionId), ServiceUrl);

        }


        #endregion
    }
}
